/*
 * Dombey Copyright (c) 2010-2023, James Bailie.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Multi-process SCGI application server as a library.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/event.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <sys/un.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>

#include <pwd.h>
#include <grp.h>
#include <signal.h>

#include <syslog.h>
#include <stdarg.h>
#include <string.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/rand.h>

#define SECOND      1000000000
#define NANOSECONDS 500000000

#define NOT_BUSY 0
#define BUSY 1
#define EMPTY -1

#define MAX_ENV 50
#define MAX_PARAM 100
#define MAX_COOKIE 50
#define MAX_FILE 250

#define BUFSIZE 262144

extern char *optarg;

void scgi_set_sigterm_intr();

EVP_CIPHER_CTX *scgi_ctx;

int scgi_fd = -1, scgi_testing = 0, scgi_backlog = 1024, scgi_min_workers = 5,
   scgi_max_body = 10000, scgi_logging = 0, scgi_inlen = 0, scgi_context = 0,
   scgi_max_file = 2000000, scgi_worker_no = 0, scgi_header_len = 0,
   scgi_interval = 0, scgi_counter = 0, scgi_encrypting = 0, scgi_cbc = 0,
   scgi_read_eof = 0, scgi_max_workers = 25, scgi_max_form = MAX_PARAM;

void ( *scgi_periodic_func )() = NULL;

volatile int scgi_killed = 0;

pid_t *scgi_workers = NULL;
FILE *scgi_conn = NULL;

char *scgi_interface = "", *scgi_port = "4000", *scgi_group = "nobody",
   *scgi_user = "nobody", *scgi_busymap = NULL, *scgi_header = NULL,
   *scgi_body = NULL, *scgi_root_dir = NULL, *scgi_listen_unix = NULL,

   *scgi_env[ MAX_ENV + 1 ], *scgi_params[ MAX_PARAM + 1 ],
   *scgi_cookies[ MAX_COOKIE + 1 ], *scgi_files[ MAX_FILE + 1 ],

   *scgi_config_file = NULL, *scgi_hostname = NULL,
   *scgi_query_string = NULL, *scgi_http_cookie = NULL,
   *scgi_app_name = "dombey", *scgi_pidfile = "/var/run/dombey.pid";

unsigned char scgi_key[ 32 ], scgi_iv[ 16 ];

struct passwd *scgi_passwd;
struct group *scgi_grp;

char *scgi_get_env( char * );
char **scgi_get_envs();

char *scgi_get_param( char * );
char **scgi_get_params();

char *scgi_get_cookie( char * );
char **scgi_get_cookies();

void scgi_set_key_iv( char *, char * );
void scgi_set_name( char * );

void scgi_init_func();
void scgi_exit_func();
void scgi_worker_init_func();

void scgi_request_handler( FILE * );
char *scgi_form_encode( char * );

void scgi_process_params();
void scgi_process_cookies();

int scgi_stack_inc = 128;

struct scgi_string
{
   int free, used;
   char *top;
   char *str;
};

void scgi_log_error()
{
   unsigned long err;

   while(( err = ERR_get_error()))
   {
      char *str = ERR_error_string( err, NULL );

      if ( scgi_logging )
         syslog( LOG_ERR, "OpenSSL: %s", str );
      else
         fprintf( stderr, "OpenSSL: %s\n", str );
   }
}

void *scgi_memory( int size )
{
   void *ptr;

   if ( size == 0 )
      return NULL;

   if (( ptr = malloc( size )) == NULL )
   {
      if ( scgi_logging )
         syslog( LOG_WARNING, "malloc(): %m" );
      else
         fprintf( stderr, "malloc(): %s\n", strerror( errno ));

      exit( 1 );
   }

   return ptr;
}

char *scgi_str_dup( char *str, int len )
{
   char *ptr;

   if ( len < 0 )
      for( len = 0, ptr = str; *ptr; ++ptr )
         ++len;

   ptr = ( char *)scgi_memory( len + 1 );
   bcopy( str, ptr, len );
   ptr[ len ] = '\0';

   return ptr;
}

struct scgi_string *scgi_make_string()
{
   struct scgi_string *s;

   s = ( struct scgi_string *)scgi_memory( sizeof( struct scgi_string ));
   s->str = ( char *)scgi_memory( scgi_stack_inc + 1 );
   *s->str = '\0';
   s->free = scgi_stack_inc;
   s->used = 0;
   s->top = s->str;

   return s;
}

void scgi_string_append( struct scgi_string *s, char c )
{
   if ( s->free == 0 )
   {
      s->str = ( char *)realloc( s->str, s->used + 1 + scgi_stack_inc );

      if ( s->str == NULL )
      {
         if ( scgi_logging )
            syslog( LOG_WARNING, "realloc(): %m" );
         else
            fprintf( stderr, "realloc(): %s\n", strerror( errno ));

         exit( 1 );
      }

      /* Leave room for end-of-string sentinel */

      s->free = scgi_stack_inc;
      s->top = &s->str[ s->used ];
   }

   ++s->used;
   --s->free;
   *s->top++ = c;
   *s->top = '\0';
}

#define STRING_APPEND( _s_, _c_ ) { if (( _s_ )->free )\
   {\
      --( _s_ )->free;\
      ++( _s_ )->used;\
      *( _s_ )->top++ = _c_;\
      *( _s_ )->top ='\0';\
   }\
   else scgi_string_append( _s_, _c_ ); }

void scgi_string_truncate( struct scgi_string *s )
{
   if ( s->used )
   {
      s->free += s->used;
      s->used = 0;
      *s->str = '\0';
      s->top = s->str;
   }
}

char *scgi_html_escape( char *str )
{
   static struct scgi_string *s = NULL;
   char *ptr;

   if ( s == NULL )
      s = scgi_make_string();
   else
      scgi_string_truncate( s );

   for( ptr = str; *ptr; ++ptr )
      switch( *ptr )
      {
         case '&':
            STRING_APPEND( s, '&' )
            STRING_APPEND( s, 'a' )
            STRING_APPEND( s, 'm' )
            STRING_APPEND( s, 'p' )
            STRING_APPEND( s, ';' )
            break;

         case '<':
            STRING_APPEND( s, '&' )
            STRING_APPEND( s, 'l' )
            STRING_APPEND( s, 't' )
            STRING_APPEND( s, ';' )
            break;

         case '>':
            STRING_APPEND( s, '&' )
            STRING_APPEND( s, 'g' )
            STRING_APPEND( s, 't' )
            STRING_APPEND( s, ';' )
            break;

         default:
            STRING_APPEND( s, *ptr )
      }

   return scgi_str_dup( s->str, s->used );
}

char *scgi_form_decode( char *str )
{
   static struct scgi_string *s = NULL;
   char *ptr;

   if ( s == NULL )
      s = scgi_make_string();
   else
      scgi_string_truncate( s );

   for( ptr = str; *ptr; ++ptr )
      switch( *ptr )
      {
         case '+':
            STRING_APPEND( s, ' ' )
            break;

         case '%':
            if ( ! *( ptr + 1 ) || ! *( ptr + 2 ))
               STRING_APPEND( s, *ptr )
            else
            {
               char data[ 3 ];
               char code;

               data[ 0 ] = *( ptr + 1 );
               data[ 1 ] = *( ptr + 2 );
               data[ 2 ] = '\0';

               if (( code = ( char )strtol( data, NULL, 16 )))
               {
                  STRING_APPEND( s, code )
                  ptr += 2;
               }
               else
                  STRING_APPEND( s, *ptr )
            }
            break;

         default:
            STRING_APPEND( s, *ptr )
      }

   return scgi_str_dup( s->str, s->used );
}

char *scgi_form_encode( char *str )
{
   char *ptr, *ptr2;
   static struct scgi_string *s = NULL;
   int found;
   static char *reserved = "']$;:@&+!?=#{}/[^`~\"<>|%\\\t\r\n",
               *letters  = "0123456789ABCDEF";

   if ( s == NULL )
      s = scgi_make_string();
   else
      scgi_string_truncate( s );

   for( ptr = str; *ptr; ++ptr )
   {
      found = 0;

      if ( *ptr == ' ' )
      {
         STRING_APPEND( s, '+' )
         continue;
      }

      if ( *ptr < 32 || *ptr > 126 )
      {
         STRING_APPEND( s, '%' )
         STRING_APPEND( s, letters[ *ptr / 16 ] )
         STRING_APPEND( s, letters[ *ptr % 16 ] )
         continue;
      }

      for( ptr2 = reserved; *ptr2; ++ptr2 )
         if ( *ptr == *ptr2 )
         {
            STRING_APPEND( s, '%' )
            STRING_APPEND( s, letters[ *ptr / 16 ] )
            STRING_APPEND( s, letters[ *ptr % 16 ] )
            found = 1;
            break;
         }

      if ( ! found )
         STRING_APPEND( s, *ptr )
   }

   return scgi_str_dup( s->str, s->used );
}

void scgi_set_hostname( char *name )
{
   long len;
   char *ptr1, *ptr2;

   if (( len = sysconf( _SC_HOST_NAME_MAX )) < 0 )
      len = MAXPATHLEN;

   ++len;
   scgi_hostname = scgi_memory( len );

   if ( name != NULL )
   {
      ptr1 = scgi_hostname;
      ptr2 = name;
      --len;

      while( len && *ptr2 )
      {
         *ptr1++ = *ptr2++;
         --len;
      }

      *ptr1 = '\0';
      return;
   }

   if ( gethostname( scgi_hostname, len ))
   {
      fprintf( stderr, "%s: gethostname(): %s", scgi_app_name, strerror( errno ));
      exit( 1 );
   }

   ptr1 = "localhost";
   ptr2 = scgi_hostname;

   while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
   {
      ++ptr1;
      ++ptr2;
   }

   if ( ! *ptr1 && ! *ptr2 )
   {
      ptr1 = scgi_hostname;
      ptr2 = "127.0.0.1";
      --len;

      while( len && *ptr2 )
      {
         *ptr1++ = *ptr2++;
         --len;
      }

      *ptr1 = '\0';
   }
}

int scgi_check_header_origin( char *header, char *hostname )
{
   char *ptr1, *ptr2, *ptr3, *ptr4;
   int tries;

   ptr1 = "https://";
   ptr2 = header;

   while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
   {
      ++ptr1;
      ++ptr2;
   }

   if ( *ptr1 )
      return 1;

   if ( *ptr2 == '[' )
      ++ptr2;

   ptr1 = hostname;
   ptr4 = ptr2;
   tries = 0;

AGAIN:
   while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
   {
      ++ptr1;
      ++ptr2;
   }

   if ( ! *ptr1 && ( ! *ptr2 || *ptr2 == '/' || *ptr2 == ']' ))
      return 0;

   if ( ++tries == 2 )
      return 1;

   ptr1 = hostname;
   ptr3 = "::ffff:";

   while( *ptr1 && *ptr3 && *ptr1 == *ptr3 )
   {
      ++ptr1;
      ++ptr3;
   }

   if ( ! *ptr3 )
   {
      ptr2 = ptr4;
      goto AGAIN;
   }

   return 1;
}

int scgi_check_origin()
{
   char *header;

   if ( scgi_hostname == NULL ||
        (( header = scgi_get_env( "HTTP_ORIGIN" )) == NULL &&
         ( header = scgi_get_env( "HTTP_REFERER" )) == NULL &&
         ( header = scgi_get_env( "HTTP_REFERRER" )) == NULL ))
      return 1;

   if ( scgi_check_header_origin( header, scgi_hostname ))
   {
      char *addr;

      if (( addr = scgi_get_env( "LOCAL_ADDR" )) != NULL )
         return scgi_check_header_origin( header, addr );

      return 1;
   }

   return 0;
}

/*
 * If we are a worker process, we set the killed flag, and when we are
 * finished servicing any current connection, we will exit.  It would be
 * nice to unlink the master process's pidfile here, but we no longer are
 * root and so cannot access /var/run/
 */

void scgi_signal_handler( int signo )
{
   scgi_killed = 1;

   if ( scgi_testing )
   {
      scgi_exit_func();
      exit( 0 );
   }
}

void scgi_kill_workers_and_exit()
{
   int i, alive;
   pid_t pid;
   struct timespec tm;

   tm.tv_sec = 1;
   tm.tv_nsec = 0;

   for( ; ; )
   {
      for( alive = 0, i = 0; i < scgi_max_workers; ++i )
         if ( scgi_workers[ i ] )
         {
            ++alive;
            kill( scgi_workers[ i ], SIGTERM );
         }

      if ( ! alive )
         break;

      nanosleep( &tm, NULL );

      while(( pid = waitpid( -1, NULL, WNOHANG )) > 0 )
      {
         for( i = 0; i < scgi_max_workers; ++i )
            if ( pid == scgi_workers[ i ] )
            {
               scgi_workers[ i ] = 0;
               scgi_busymap[ i ] = EMPTY;
               break;
            }
      }
   }

   scgi_exit_func();
   exit( 0 );
}

/*
 * x - Do not become daemon.  Run single process iterative server.

 * r - root directory
 * l - socket for unix domain server
 * m - max workers
 * n - min free workers
 * p - port
 * f - config file path
 * i - interface
 * u - user
 * g - group
 * b - max size of request bodies
 * e - max size of uploaded files
 * f - config file
 */

void scgi_set_options( int argc, char **argv )
{
   int i, specified = 0;
   char *ptr;

   while(( i = getopt( argc, argv, "xca:q:e:l:f:r:m:b:n:p:i:u:g:")) != -1 )
      switch( i )
      {
         case 'c':
            ++scgi_cbc;
            break;

         case 'q':
            scgi_backlog = strtol( optarg, NULL, 10 );
            break;

         case 'a':
            scgi_max_form = strtol( optarg, NULL, 10 ) * 2;
            break;

         case 'b':
            scgi_max_body = strtol( optarg, NULL, 10 );
            break;

         case 'e':
            scgi_max_file = strtol( optarg, NULL, 10 );
            break;

         case 'f':
            scgi_config_file = optarg;
            break;

         case 'r':
            scgi_root_dir = optarg;
            break;

         case 'g':
            scgi_group = optarg;
            break;

         case 'i':
            ++specified;
            scgi_interface = optarg;
            break;

         case 'l':
            scgi_listen_unix = optarg;
            break;

         case 'm':
            scgi_max_workers = strtol( optarg, NULL, 10 );
            break;

         case 'n':
            scgi_min_workers = strtol( optarg, NULL, 10 );
            break;

         case 'p':
            ++specified;
            scgi_port = optarg;
            break;

         case 'u':
            scgi_user = optarg;
            break;

         case 'x':
            scgi_testing = 1;
      }

   if ( specified && scgi_listen_unix != NULL )
   {
      fprintf( stderr,
               "%s: the -l option cannot be present when either or both of the -i and -p options are present\n",
               scgi_app_name );
      exit( 1 );
   }

   if ( scgi_backlog < 1024 )
   {
      fprintf( stderr, "%s: ignoring -q option value < 1024: %d\n", scgi_app_name, scgi_backlog );
      scgi_backlog = 1024;
   }

   if ( scgi_max_form > MAX_PARAM || scgi_max_form < 2 )
   {
      fprintf( stderr, "%s: the value of the -a option must be in the range 1-50.\n", scgi_app_name );
      exit( 1 );
   }

   if ( scgi_root_dir == NULL || ! *scgi_root_dir )
   {
      fprintf( stderr, "%s: root directory undefined (-r option)\n", scgi_app_name );
      exit( 1 );
   }

   if ( chdir( scgi_root_dir ) < 0 )
   {
      fprintf( stderr, "%s: chdir( %s ): %s\n", scgi_app_name, scgi_root_dir, strerror( errno ));
      exit( 1 );
   }

   /*
    * Remove trailing / but allow root directory to be scgi_root_dir.
    */

   for( ptr = scgi_root_dir; *ptr; ++ptr )
      ++ptr;

   for( --ptr; *ptr == '/' && ptr > scgi_root_dir; --ptr )
      ;

   if ( *++ptr == '/' )
      *ptr = '\0';

   if ( scgi_min_workers <= 0 )
   {
      fprintf( stderr, "%s: min workers <= 0: %d\n", scgi_app_name, scgi_min_workers );
      exit( 1 );
   }

   if ( scgi_max_workers <= 0 )
   {
      fprintf( stderr, "%s: max workers <= 0: %d\n", scgi_app_name, scgi_max_workers );
      exit( 1 );
   }

   if ( scgi_min_workers > scgi_max_workers )
   {
      fprintf( stderr, "%s: min workers > max_workers: %d > %d\n", scgi_app_name, scgi_min_workers, scgi_max_workers );
      exit( 1 );
   }

   if ( scgi_max_body <= 0 )
   {
      fprintf( stderr, "%s: max_body <= 0: %d\n", scgi_app_name, scgi_max_body );
      exit( 1 );
   }

   if (( scgi_passwd = getpwnam( scgi_user )) == NULL )
   {
      fprintf( stderr, "%s: user \"%s\" does not exist\n", scgi_app_name, scgi_user );
      exit( 1 );
   }

   if (( scgi_grp = getgrnam( scgi_group )) == NULL )
   {
      fprintf( stderr, "%s: group \"%s\" does not exist\n", scgi_app_name, scgi_group );
      exit( 1 );
   }
}

void scgi_become_daemon()
{
   char buffer[ 16 ];
   int len, file;

   /*
    * Fork and let the parent die, continuing as child so we are not a
    * process group leader.  This is necessary for the call to setsid().
    */

   switch( fork() )
   {
      case -1:
         fprintf( stderr, "%s: fork(): %s\n", scgi_app_name, strerror( errno ));
         exit( 1 );

      case 0:
         break;

      default:
         exit( 0 );
   }

   fclose( stdout );
   fclose( stderr );
   fclose( stdin );

   stdin = fopen( "/dev/null", "r" );
   stdout = fopen( "/dev/null", "w" );
   stderr = fopen( "/dev/null", "w" );

   if ( stdin == NULL || stdout == NULL || stderr == NULL )
   {
      syslog( LOG_ERR, "fopen(): %m" );
      exit( 1 );
   }

   /*
    * Detach us from our controlling terminal, so job control and other
    * signals may not be sent to us from that terminal.
    */

   if ( setsid() < 0 )
   {
      syslog( LOG_ERR, "setsid(): %m" );
      exit( 1 );
   }

   /*
    * Write our pid to disk.
    */

   if (( file = open( scgi_pidfile, O_WRONLY | O_CREAT | O_TRUNC, S_IWUSR | S_IRUSR | S_IRGRP )) < 0 )
   {
      syslog( LOG_WARNING, "open(): %m" );
      return;
   }

   len = snprintf( buffer, sizeof( buffer ), "%d", getpid() );
   write( file, buffer, len );
   close( file );

   /*
    * We don't fork children which create files, but if we did, this will
    * make those files not world-accessible by default.
    */

   umask( 7 );
}

void scgi_change_identity()
{
   if ( setgid( scgi_grp->gr_gid ) < 0 )
      syslog( LOG_ERR, "setgid(): %m" );

   if ( setuid( scgi_passwd->pw_uid ) < 0 )
      syslog( LOG_ERR, "setuid(): %m" );
}

void scgi_close_on_exec( int fd )
{
   if ( fcntl( fd, F_SETFD, FD_CLOEXEC ) < 0 )
      syslog( LOG_ERR, "fcntl( F_SETFD, FD_CLOEXEC ): %m" );
}

void scgi_start_listening_unix()
{
   struct sockaddr_un sa;

   if (( scgi_fd = socket( PF_LOCAL, SOCK_STREAM, 0 )) < 0 )
   {
      syslog( LOG_ERR, "socket(): %m" );
      exit( 1 );
   }

   unlink( scgi_listen_unix );
   bzero( &sa, sizeof( struct sockaddr_un ));
   sa.sun_family = AF_UNIX;

   {
      char *ptr1, *ptr2;
      int len;

      ptr1 = sa.sun_path;
      ptr2 = scgi_listen_unix;
      len = sizeof( sa.sun_path ) - 1;

      while( len && *ptr2 )
      {
         *ptr1++ = *ptr2++;
         --len;
      }
   }

   if ( bind( scgi_fd, ( struct sockaddr *)&sa, SUN_LEN( &sa )))
   {
      syslog( LOG_ERR, "bind( %s ): %m", scgi_listen_unix );
      close( scgi_fd );
      exit( 1 );
   }

   if ( chown( scgi_listen_unix, scgi_passwd->pw_uid, scgi_grp->gr_gid ) < 0 )
   {
      syslog( LOG_ERR, "chown( %s ): %m", scgi_listen_unix );
      close( scgi_fd );
      exit( 1 );
   }

   if ( chmod( scgi_listen_unix, S_IRWXU | S_IRWXG ) < 0 )
   {
      syslog( LOG_ERR, "chmod( %s, S_IRWXU | S_IRWXG ): %m", scgi_listen_unix );
      close( scgi_fd );
      exit( 1 );
   }

   if ( listen( scgi_fd, scgi_backlog ) < 0 )
   {
      syslog( LOG_ERR, "listen(): %m" );
      close( scgi_fd );
      exit( 1 );
   }

   scgi_close_on_exec( scgi_fd );
}

void scgi_start_listening()
{
   struct addrinfo hints, *res;
   int result;

   if ( scgi_listen_unix != NULL )
   {
      scgi_start_listening_unix();
      return;
   }

   /*
    * This is the new, protocol-independent way of setting up a listening
    * socket.
    */

   bzero( &hints, sizeof( struct addrinfo ));
   hints.ai_flags = AI_PASSIVE;
   hints.ai_socktype = SOCK_STREAM;

   /*
    * If the user has not specified an interface, we listen on the
    * IPv6 wildcard address.
    */

   if (( result = getaddrinfo( ( *scgi_interface ? scgi_interface : NULL ), scgi_port, &hints, &res )))
   {
      syslog( LOG_ERR, "getaddrinfo(): %s", gai_strerror( result ));
      exit( 1 );
   }

   if ( res == NULL )
   {
      syslog( LOG_ERR, "getaddrinfo(): no interface found" );
      exit( 1 );
   }

   scgi_fd = socket( res->ai_family, res->ai_socktype, res->ai_protocol );

   if ( scgi_fd == -1 )
   {
      syslog( LOG_ERR, "socket(): %m" );
      exit( 1 );
   }

   result = 1;

   /*
    * Allow duplicate bindings, so we can have more than one
    * server process, if necessary.
    */

   if ( setsockopt( scgi_fd, SOL_SOCKET, SO_REUSEPORT, &result, sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_REUSEPORT ): %m" );

   /*
    * Try and detect connections which go idle for long periods of time.
    */

   if ( setsockopt( scgi_fd, SOL_SOCKET, SO_KEEPALIVE, &result, sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_KEEPALIVE ): %m" );

   /*
    * Make sure we can accept IPv4 traffic as well as IPv6, if we
    * are bound to the IPv6 wildcard address.  We are turning off
    * the option because result = 0.
    */

   result = 0;

   if ( ! *scgi_interface &&
        setsockopt( scgi_fd, IPPROTO_IPV6, IPV6_BINDV6ONLY, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( IPV6_BINDV6ONLY ): %m" );

   if ( bind( scgi_fd, res->ai_addr, res->ai_addrlen ) < 0 )
   {
      syslog( LOG_ERR, "bind(): %m" );
      exit( 1 );
   }

   if ( res != NULL )
      freeaddrinfo( res );

   if ( listen( scgi_fd, scgi_backlog ) < 0 )
   {
      syslog( LOG_ERR, "listen(): %m" );
      exit( 1 );
   }

   scgi_close_on_exec( scgi_fd );
}

/*
 * If we are a worker, and we receive a SIGTERM, we will exit at our next
 * convenience, either the next time accept_connection() is called, after
 * we have closed the previous connection, but before we accept another, or
 * if we are already blocked in accept(), immediately, when accept() is
 * interrupted by the signal.
 */

void scgi_accept_connection()
{
   int conn, err;

AGAIN:
   if ( scgi_killed )
      exit( 0 );

   if ( scgi_conn != NULL )
   {
      fclose( scgi_conn );
      scgi_conn = NULL;
   }

   scgi_set_sigterm_intr();
   scgi_busymap[ scgi_worker_no ] = NOT_BUSY;

   conn = accept( scgi_fd, NULL, 0 );
   err = errno;

   scgi_busymap[ scgi_worker_no ] = BUSY;
   signal( SIGTERM, scgi_signal_handler );

   if ( conn < 0 )
   {
      if ( err == EINTR )
         goto AGAIN;

      syslog( LOG_ERR, "accept(): %s", strerror( err ));
      exit( 1 );
   }

   scgi_close_on_exec( conn );

   if (( scgi_conn = fdopen( conn, "r+" )) == NULL )
   {
      syslog( LOG_ERR, "fdopen(): %m" );
      exit( 1 );
   }
}

char *scgi_downcase( char *str, char stp )
{
   char *ptr;

   for( ptr = str; *ptr; ++ptr )
      if ( stp && *ptr == stp )
         break;
      else if ( *ptr >= 65 && *ptr <= 90 )
         *ptr += 32;

   return str;
}

int scgi_match_strings( char **strings, char *line, int whitespace, char **next )
{
   char **ptr, *ptr1, *ptr2;

   ptr2 = line;

   for( ptr = strings; *ptr != NULL; ++ptr )
   {
      ptr1 = *ptr;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( *ptr1 )
         return 0;

      if ( whitespace )
         while( *ptr2 == ',' || *ptr2 == ';' || *ptr2 == ' ' || *ptr2 == '\t' || *ptr2 == '\r' || *ptr2 == '\n' )
            ++ptr2;
   }

   if ( next != NULL )
      *next = ptr2;

   return 1;
}

int scgi_is_boundary( char *line, char *boundary )
{
   char *ptr1, *ptr2, *dashes = "--";

   ptr1 = dashes;
   ptr2 = line;

   while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
   {
      ++ptr1;
      ++ptr2;
   }

   if ( ! *ptr1 )
   {
      ptr1 = boundary;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( ! *ptr1 )
      {
         ptr1 = dashes;

         while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
         {
            ++ptr1;
            ++ptr2;
         }

         if ( ! *ptr1 )
            return -1;

         return 1;
      }
   }

   return 0;
}

int scgi_advance_past_boundary( char *boundary )
{
   char buffer[ 1024 ], *line;
   int bound;

   do
   {
      if (( line = fgets( buffer, sizeof( buffer ), scgi_conn )) == NULL )
      {
         syslog( LOG_ERR, ( ferror( scgi_conn ) ? "fgets(): %m" : "first multipart boundary not found" ));
         return 1;
      }
   }
   while( ! ( bound = scgi_is_boundary( line, boundary )));

   if ( bound < 0 )
   {
      syslog( LOG_ERR, "empty form-data document" );
      return 1;
   }

   return 0;
}

char *scgi_extract_entity_param( char *line, char *param, char **start )
{
   char *ptr1, *ptr2;
   int len;

   *start = NULL;

   ptr1 = param;
   ptr2 = scgi_downcase( line, '=' );

   while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
   {
      ++ptr1;
      ++ptr2;
   }

   if ( *ptr1 )
      return NULL;

   len = 0;

   if ( *ptr2 == '"' )
   {
      ptr1 = ++ptr2;

      while( *ptr2 && *ptr2 != '"' )
      {
         ++ptr2;
         ++len;
      }
   }
   else
   {
      ptr1 = ptr2;

      while( *ptr2 && *ptr2 != ',' && *ptr2 != ';' && *ptr2 != ' ' && *ptr2 != '\t' && *ptr2 != '\r' && *ptr2 != '\n' )
      {
         ++ptr2;
         ++len;
      }
   }

   while( *ptr2 == '"' || *ptr2 == ',' || *ptr2 == ';' || *ptr2 == ' ' || *ptr2 == '\t' || *ptr2 == '\r' || *ptr2 == '\n' )
      ++ptr2;

   *start = ptr2;

   return scgi_str_dup( ptr1, len );
}

char *scgi_extract_mime_boundary( char *line )
{
   char *ptr1, *ptr2, *result;

   ptr2 = line;

AGAIN:
   while( *ptr2 == ',' || *ptr2 == ';' || *ptr2 == ' ' || *ptr2 == '\t' || *ptr2 == '\r' || *ptr2 == '\n' )
      ++ptr2;

   result = scgi_extract_entity_param( ptr2, "boundary=", &ptr1 );

   /*
    * We ran out of string or we found another unrecognized parameter.
    * Advance to the next parameter if possible.
    */

   if ( ptr1 == NULL )
   {
      while( *ptr2 && *ptr2 != ',' && *ptr2 != ';' && *ptr2 != ' ' && *ptr2 != '\t' && *ptr2 != '\r' && *ptr2 != '\n' )
         ++ptr2;

      if ( *ptr2 )
         goto AGAIN;

      syslog( LOG_ERR, "no boundary defined for multipart/form-data" );
      return NULL;
   }

   /*
    * if ptr1 != NULL, then result != NULL.
    */

   if ( ! *result )
   {
      free( result );
      syslog( LOG_ERR, "empty boundary defined for multipart/form-data" );
      return NULL;
   }

   if ( scgi_advance_past_boundary( result ))
   {
      free( result );
      return NULL;
   }

   return result;
}

char **scgi_extract_entity_names( char *line )
{
   char *ptr1, *ptr2, *strings[] = { "content-disposition:", "form-data", NULL };
   int n;
   static char *result[ 2 ];

   result[ 0 ] = NULL;
   result[ 1 ] = NULL;

   scgi_downcase( line, ':' );

   if ( ! scgi_match_strings( strings, line, 1, &ptr2 ))
      return NULL;

   /*
    * Tolerate parameters in any order mixed in with other unrecognized parameters.
    */

   for( n = 0; n < 2; )
   {
      ptr1 = NULL;

      if ( result[ 0 ] == NULL )
      {
         if (( result[ 0 ] = scgi_extract_entity_param( ptr2, "name=", &ptr1 )) != NULL )
            ++n;
      }

      if ( result[ 1 ] == NULL )
      {
         if ( ptr1 != NULL )
            ptr2 = ptr1;

         if (( result[ 1 ] = scgi_extract_entity_param( ptr2, "filename=", &ptr1 )) != NULL )
            ++n;
      }

      /*
       * We ran out of string or we found another unrecognized parameter.
       * Advance to the next parameter if possible.
       */

      if ( ptr1 != NULL )
         ptr2 = ptr1;
      else
      {
         while( *ptr2 && *ptr2 != ',' && *ptr2 != ';' && *ptr2 != ' ' && *ptr2 != '\t' && *ptr2 != '\r' && *ptr2 != '\n' )
            ++ptr2;

         while( *ptr2 == ',' || *ptr2 == ';' || *ptr2 == ' ' || *ptr2 == '\t' || *ptr2 == '\r' || *ptr2 == '\n' )
            ++ptr2;

         if ( ! *ptr2 )
            break;
      }
   }

   if ( result[ 0 ] == NULL )
   {
      if ( result[ 1 ] != NULL )
         free( result[ 1 ] );

      return NULL;
   }

   return result;
}

char *scgi_read_encrypted( int fd, int *len, char *key, char *iv )
{
   unsigned char outbuf[ BUFSIZE + EVP_MAX_BLOCK_LENGTH ], inbuf[ BUFSIZE ];
   int inlen;

   if ( fd == -2 )
   {
      if ( scgi_context )
      {
         EVP_CIPHER_CTX_cleanup( scgi_ctx );
         scgi_context = 0;
      }

      if ( scgi_ctx == NULL && ( scgi_ctx = EVP_CIPHER_CTX_new()) == NULL )
      {
         scgi_log_error();
         return NULL;
      }

      EVP_DecryptInit_ex( scgi_ctx, ( scgi_cbc ? EVP_aes_256_cbc() : EVP_aes_256_gcm()), NULL, ( unsigned char *)key, ( unsigned char *)iv );
      *len = 0;
      scgi_read_eof = 0;
      ++scgi_context;
      return NULL;
   }

   if ( fd == -1 )
   {
      EVP_CIPHER_CTX_cleanup( scgi_ctx );
      *len = 0;
      scgi_context = 0;
      return NULL;
   }

   if ( ! scgi_context )
   {
      syslog( LOG_ERR, "scgi_read_encrypted(): no context" );
      *len = -1;
      return NULL;
   }

   if (( inlen = read( fd, inbuf, BUFSIZE )) < 0 )
   {
      *len = -1;
      EVP_CIPHER_CTX_cleanup( scgi_ctx );
      syslog( LOG_ERR, "read(): %m" );
      scgi_context = 0;
      return NULL;
   }

   if ( ! inlen )
      ++scgi_read_eof;

   *len = 0;

   if ( inlen )
   {
      if ( ! EVP_DecryptUpdate( scgi_ctx, outbuf, len, ( unsigned char *)inbuf, inlen ))
      {
         *len = -1;
         EVP_CIPHER_CTX_cleanup( scgi_ctx );
         scgi_log_error();
         scgi_context = 0;
         return NULL;
      }
   }

   if ( ! *len )
   {
      if ( scgi_read_eof <= 1 )
      {
         if ( EVP_DecryptFinal_ex( scgi_ctx, outbuf, len ))
            ++scgi_read_eof;
         else
         {
            *len = -1;
            EVP_CIPHER_CTX_cleanup( scgi_ctx );
            scgi_log_error();
            scgi_context = 0;
            return NULL;
         }
      }

      if ( ! *len )
      {
         EVP_CIPHER_CTX_cleanup( scgi_ctx );
         scgi_context = 0;
         return NULL;
      }
   }

   return scgi_str_dup(( char *)outbuf, *len );
}

int scgi_fwrite( char *data, int len, FILE *file )
{
   int outlen = 0;
   unsigned char outbuf[ BUFSIZE + EVP_MAX_BLOCK_LENGTH ];

   if ( data != NULL )
   {
      if ( ! EVP_EncryptUpdate( scgi_ctx, outbuf, &outlen, ( unsigned char *)data, len ))
      {
         scgi_log_error();
         return 1;
      }
   }
   else
   {
      if ( ! EVP_EncryptFinal_ex( scgi_ctx, outbuf, &outlen ))
      {
         scgi_log_error();
         return 1;
      }
   }

   if ( outlen && ! fwrite( outbuf, outlen, 1, file ))
   {
      syslog( LOG_ERR, "fwrite(): %m" );
      return 1;
   }

   return 0;
}

void scgi_clear_strings( struct scgi_string **s, struct scgi_string **l )
{
   if ( *s == NULL )
   {
      *s = scgi_make_string();
      *l = scgi_make_string();
   }
   else
   {
      ( *s )->free += ( *s )->used;
      ( *s )->used = 0;
      ( *s )->top = ( *s )->str;

      ( *l )->free += ( *l )->used;
      ( *l )->used = 0;
      ( *l )->top = ( *l )->str;
   }
}

int scgi_init_cipher( FILE *file )
{
   if ( file != NULL && scgi_encrypting )
   {
      if ( ! RAND_bytes( scgi_key, 32 ))
      {
         scgi_log_error();
         return 0;
      }

      if ( ! RAND_bytes( scgi_iv, 16 ))
      {
         scgi_log_error();
         return 0;
      }

      if ( scgi_ctx == NULL && ( scgi_ctx = EVP_CIPHER_CTX_new()) == NULL )
      {
         scgi_log_error();
         return 0;
      }

      EVP_CIPHER_CTX_init( scgi_ctx );
      EVP_EncryptInit_ex( scgi_ctx, ( scgi_cbc ? EVP_aes_256_cbc() : EVP_aes_256_gcm()), NULL, scgi_key, scgi_iv );
      return 1;
   }

   return 0;
}

int scgi_add_to_file( struct scgi_string **s, FILE *file, int last )
{
   if ( scgi_encrypting )
   {
      if (( *s )->used && scgi_fwrite(( *s )->str, ( *s )->used, file ))
         return 1;

      if ( last && scgi_fwrite( NULL, 0, file ))
         return 1;
   }
   else if ( ( *s )->used && ! fwrite(( *s )->str, ( *s )->used, 1, file ))
   {
      syslog( LOG_ERR, "fwrite(): %m" );
      return 1;
   }

   ( *s )->free += ( *s )->used;
   ( *s )->top = ( *s )->str;
   ( *s )->used = 0;

   return 0;
}

int scgi_check_oversize( int count, FILE *file )
{
   /*
    * Account for the MIME terminators. We check before they have been removed.
    */

   if ( file == NULL )
   {
      if ( count > scgi_max_body + 2 )
         return 1;
   }
   else if ( count > scgi_max_file + 2 )
      return 1;

   return 0;
}

int scgi_process_mime_data( char *boundary, FILE *file, char **value )
{
   static struct scgi_string *s = NULL, *l = NULL;
   int c, count = 0, result = 0, n, enc;
   char *ptr;

   *value = NULL;

   scgi_clear_strings( &s, &l );
   enc = scgi_init_cipher( file );

   for( ; ; )
   {
      for( ; ; )
      {
         if (( c = fgetc( scgi_conn )) == EOF )
         {
            /*
             * Tolerate unterminated final boundary.
             */

            if (( result = scgi_is_boundary( l->str, boundary )))
               goto SKIP;

            syslog( LOG_ERR, ( ferror( scgi_conn ) ? "fgetc(): %m" : "premature EOF in multipart/form-data stream" ));
            goto BAILOUT;
         }

         STRING_APPEND( l, c );

         if ( c == '\n' || l->used == BUFSIZE )
            break;
      }

      if (( result = scgi_is_boundary( l->str, boundary )))
         break;

      count += l->used;

      if ( scgi_check_oversize( count, file ))
         goto BAILOUT;
      else
      {
         /*
          * We write the previous line to the file so that we do not
          * write the MIME terminators if the current line is the last
          * line.  We handle the last line outside of the loop.
          */

         if ( file != NULL && scgi_add_to_file( &s, file, 0 ))
            goto BAILOUT;

         for( n = 0, ptr = l->str; n < l->used; ++ptr, ++n )
            STRING_APPEND( s, *ptr )
      }

      l->free += l->used;
      l->top = l->str;
      l->used = 0;
   }

SKIP:

   /*
    * Strip the MIME terminators.
    */

   if ( s->used > 1 )
   {
      s->used -= 2;
      s->top -= 2;
      s->free += 2;
      *s->top = '\0';
   }

   if ( file == NULL )
      *value = scgi_str_dup( s->str, s->used );
   else if ( scgi_add_to_file( &s, file, 1 ))
   {
BAILOUT:
      if ( enc )
         EVP_CIPHER_CTX_cleanup( scgi_ctx );

      return 1;
   }

   if ( enc )
      EVP_CIPHER_CTX_cleanup( scgi_ctx );

   return ( result > 0 ? 0 : result );
}

int scgi_process_mime_entity( char *boundary, char **names, char **value )
{
   char path[ MAXPATHLEN + 1 ];
   int fd, result;
   FILE *file = NULL;

   if ( names[ 1 ] != NULL )
   {
      if ( ! scgi_max_file )
         return 1;

      snprintf( path, sizeof( path ), "%s/dombey-XXXXXXXXXXXXX", scgi_root_dir );

      if (( fd = mkstemp( path )) < 0 )
      {
         syslog( LOG_ERR, "mkstemp( %s ): %m", path );
         return 1;
      }

      if (( file = fdopen( fd, "w+" )) == NULL )
      {
         syslog( LOG_ERR, "fdopen(): %m" );
         close( fd );
         return 1;
      }
   }

   result = scgi_process_mime_data( boundary, file, value );

   if ( result <= 0 && *value == NULL && file != NULL )
      *value = scgi_str_dup( path, -1 );

   if ( file != NULL )
   {
      fclose( file );

      if ( result > 0 && unlink( path ) < 0 )
         syslog( LOG_ERR, "unlink(): %m" );
   }

   return result;
}

void scgi_free_names( char **names )
{
   if ( names != NULL )
   {
      free( names[ 0 ] );

      if ( names[ 1 ] != NULL )
         free( names[ 1 ] );
   }
}

char **scgi_get_entity_names()
{
   char buffer[ 1024 ], **names = NULL;

   for( ; ; )
   {
      if ( fgets( buffer, sizeof( buffer ), scgi_conn ) == NULL )
      {
         syslog( LOG_ERR, ( ferror( scgi_conn ) ? "fgets(): %m" : "premature EOF for multipart/form-data stream" ));
         scgi_free_names( names );
         return NULL;
      }

      if ( *buffer == '\r' || *buffer == '\n' )
         break;

      if ( names == NULL )
         names = scgi_extract_entity_names( buffer );
   }

   if ( names == NULL || ! *names[ 0 ] )
   {
      scgi_free_names( names  );
      syslog( LOG_ERR, "missing or empty name parameter in multipart/form-data entity" );
      return NULL;
   }

   /*
    * An empty filename parameter from a file input indicates no file was
    * selected for uploading.  Removing the empty parameter will cause this
    * entity to be processed into the params array instead of the files
    * array.  scgi_process_mime_data() will set the value to the empty
    * string because there will be no data in the entity body.
    */

   if ( names[ 1 ] != NULL && ! *names[ 1 ] )
   {
      free( names[ 1 ] );
      names[ 1 ] = NULL;
   }

   return names;
}

int scgi_process_form_entity( char *boundary, char **name, char **value, char **file )
{
   char **names, *myval;
   int result;

   if (( names = scgi_get_entity_names()) == NULL )
      return 1;

   if (( result = scgi_process_mime_entity( boundary, names, &myval )) == 1 )
   {
      scgi_free_names( names );
      return 1;
   }

   *name = names[ 0 ];

   if ( names[ 1 ] == NULL )
      *value = myval;
   else
   {
      *value = names[ 1 ];
      *file = myval;
   }

   return result;
}

void scgi_free_data()
{
   char **ptr;

   for( ptr = scgi_params; *ptr != NULL; ++ptr )
   {
      free( *ptr );
      *ptr = NULL;
   }

   for( ptr = scgi_cookies; *ptr != NULL; ++ptr )
      *ptr = NULL;

   for( ptr = scgi_files; *ptr != NULL; ptr += 5 )
   {
      free( *ptr );
      *ptr = NULL;

      free( *( ptr + 1 ));
      *( ptr + 1 ) = NULL;

      unlink( *( ptr + 2 ));
      free( *( ptr + 2 ));
      *( ptr + 2 ) = NULL;

      if ( *( ptr + 3 ) != NULL )
      {
         free( *( ptr + 3 ));
         *( ptr + 3 ) = NULL;
      }

      if ( *( ptr + 4 ) != NULL )
      {
         free( *( ptr + 4 ));
         *( ptr + 4 ) = NULL;
      }
   }

   if ( scgi_header != NULL )
      free( scgi_header );

   if ( scgi_query_string != NULL )
      free( scgi_query_string );

   if ( scgi_http_cookie != NULL )
      free( scgi_http_cookie );

   if ( scgi_body != NULL )
      free( scgi_body );

   scgi_header = scgi_body = scgi_query_string = scgi_http_cookie = NULL;
}

int scgi_process_form_data( char *content_type )
{
   char *boundary, *name = NULL, *value = NULL, *file = NULL, **p_ptr, **f_ptr;
   int result, count;

   if (( boundary = scgi_extract_mime_boundary( content_type )) == NULL )
      return 1;

   for( count = 0, p_ptr = scgi_params; *p_ptr != NULL; ++p_ptr )
      ++count;

   f_ptr = scgi_files;

   while(( result = scgi_process_form_entity( boundary, &name, &value, &file )) <= 0 )
   {
      if (( count += 2 ) > scgi_max_form )
      {
         if ( name != NULL )
            free( name );

         if ( value != NULL )
            free( value );

         if ( file != NULL )
         {
            unlink( file );
            free( file );
         }

         if ( result < 0 )
            break;
         else
            continue;
      }

      if ( file != NULL )
      {
         *f_ptr++ = name;
         *f_ptr++ = value;
         *f_ptr++ = file;

         *f_ptr++ = ( scgi_encrypting ? scgi_str_dup(( char *)scgi_key, 32 ) : NULL );
         *f_ptr++ = ( scgi_encrypting ? scgi_str_dup(( char *)scgi_iv, 16 ) : NULL );
      }
      else
      {
         *p_ptr++ = name;
         *p_ptr++ = value;
      }

      if ( result < 0 )
         break;

      name = value = file = NULL;
   }

   free( boundary );

   if ( result > 0 )
      scgi_free_data();

   return ( result < 0 ? 0 : 1 );
}

char *scgi_find_next( char *ptr )
{
   while( --scgi_header_len )
      if ( ! *ptr++ )
         break;

   if ( ! scgi_header_len )
      return NULL;

   return ptr;
}

int scgi_init_env()
{
   char *ptr;
   int n;

   n = 0;

   for( ptr = scgi_header; ptr != NULL; ptr = scgi_find_next( ptr ))
   {
      scgi_env[ n++ ] = ptr;

      if ( n == MAX_ENV )
         break;
   }

   if ( n % 2 )
   {
      free( scgi_header );
      scgi_header = NULL;
      return 1;
   }

   scgi_env[ n ] = NULL;
   scgi_process_cookies();

   return 0;
}

char **scgi_lookup_list( char **list, char *key )
{
   char **ptr, *ptr1, *ptr2;

   if ( list == NULL )
      return NULL;

   for( ptr = list; *ptr != NULL; ptr += ( list == scgi_files ? 5 : 2 ))
   {
      ptr1 = *ptr;
      ptr2 = key;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( ! *ptr1 && ! *ptr2 )
         return &ptr[ 1 ];
   }

   return NULL;
}

char *scgi_get_env( char *key )
{
   char **ptr;

   if (( ptr = scgi_lookup_list( scgi_env, key )) == NULL )
      return NULL;

   return *ptr;
}

char **scgi_get_envs()
{
   return scgi_env;
}

int scgi_read_header()
{
   char *ptr, buffer[ 64 ];
   int c, i;

   for( i = 0, ptr = buffer; i < sizeof( buffer ); ++i, ++ptr )
   {
      if (( c = fgetc( scgi_conn )) == EOF )
         return 1;

      if ( c == ':' )
      {
         *ptr = '\0';
         break;
      }

      *ptr = c;
   }

   if ( c != ':' )
      return 1;

   if (( scgi_header_len = strtol( buffer, NULL, 10 )) <= 0 )
      return 1;

   scgi_header = scgi_memory( scgi_header_len );

   if ( fread( scgi_header, 1, scgi_header_len, scgi_conn ) < scgi_header_len )
   {
      free( scgi_header );
      scgi_header = NULL;
      return 1;
   }

   if ( fgetc( scgi_conn ) != ',' )
   {
      free( scgi_header );
      scgi_header = NULL;
      return 1;
   }

   return scgi_init_env();
}

void scgi_process_cookies()
{
   char *cookie, *name, *ptr;
   int i;

   if (( cookie = scgi_get_env( "HTTP_COOKIE" )) == NULL || ! *cookie )
      return;

   scgi_http_cookie = scgi_str_dup( cookie, -1 );
   cookie = scgi_http_cookie;

   for( i = 0, ptr = cookie; *ptr; ++ptr )
   {
      if ( *ptr == ';' || *ptr == ',' )
      {
         if ( i >= MAX_COOKIE )
            break;

         *ptr = '\0';

         for( name = cookie; *name && *name != '='; ++name )
            ;

         if ( *name )
            *name++ = '\0';

         while( *cookie == ' ' )
            ++cookie;

         scgi_cookies[ i++ ] = cookie;
         scgi_cookies[ i++ ] = name;

         cookie = ptr + 1;
      }
   }

  if ( *cookie && i < MAX_COOKIE - 1 )
  {
      for( name = cookie; *name && *name != '='; ++name )
         ;

      if ( *name )
         *name++ = '\0';

      while( *cookie == ' ' )
         ++cookie;

      scgi_cookies[ i++ ] = cookie;
      scgi_cookies[ i++ ] = name;
   }
}

void scgi_process_params()
{
   char *query, *ptr, *param;
   int i, flag = 0;

   if (( query = scgi_get_env( "QUERY_STRING" )) == NULL || ! *query )
   {
      if (( query = scgi_body ) == NULL || ! *query )
         return;
   }
   else
   {
      ++flag;
      scgi_query_string = scgi_str_dup( query, -1 );
      query = scgi_query_string;
   }

   i = 0;

MORE:
   for( ptr = query; *ptr; ++ptr )
   {
      if ( *ptr == '&' )
      {
         if ( i >= MAX_PARAM )
            break;

         *ptr = '\0';

         for( param = query; *param && *param != '='; ++param )
            ;

         if ( *param )
            *param++ = '\0';

         scgi_params[ i++ ] = scgi_form_decode( query );
         scgi_params[ i++ ] = scgi_form_decode( param );

         query = ptr + 1;
      }
   }

  if ( *query && i < MAX_PARAM - 1 )
  {
      for( param = query; *param && *param != '='; ++param )
         ;

      if ( *param )
         *param++ = '\0';

      scgi_params[ i++ ] = scgi_form_decode( query );
      scgi_params[ i++ ] = scgi_form_decode( param );
   }

   if ( flag && scgi_body != NULL )
   {
      query = scgi_body;
      --flag;
      goto MORE;
   }
}

int scgi_read_body()
{
   char *ptr, *ptr1, *ptr2;
   int count;

   if (( ptr = scgi_get_env( "CONTENT_LENGTH" )) == NULL )
   {
      scgi_free_data();
      return 1;
   }

   if (( count = strtol( ptr, NULL, 10 )) < 0 )
   {
      scgi_free_data();
      return 1;
   }

   if (( ptr = scgi_get_env( "CONTENT_TYPE" )) == NULL )
   {
      scgi_process_params();
      return 0;
   }

   ptr1 = "multipart/form-data";
   ptr2 = ptr;

   while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
   {
      ++ptr1;
      ++ptr2;
   }

   if ( ! *ptr1 )
   {
      scgi_process_params();

      if ( ! count )
         return 0;

      return scgi_process_form_data( ptr2 );
   }

   if ( count > scgi_max_body )
   {
      scgi_free_data();
      return 1;
   }

   if ( ! count )
   {
      scgi_process_params();
      return 0;
   }

   scgi_body = scgi_memory( count + 1 );

   if ( fread( scgi_body, 1, count, scgi_conn ) < count )
   {
      scgi_free_data();
      return 1;
   }

   scgi_body[ count ] = '\0';
   scgi_process_params();

   return 0;
}

char *scgi_get_param( char *param )
{
   char **ptr;

   if (( ptr = scgi_lookup_list( scgi_params, param )) == NULL )
      return NULL;

   return *ptr;
}

char **scgi_get_params()
{
   return scgi_params;
}

char **scgi_get_file( char *file )
{
   return scgi_lookup_list( scgi_files, file );
}

char **scgi_get_next_file( char **last, char *file )
{
   char **ptr, *ptr1, *ptr2;

   if ( last == NULL || file == NULL )
      return NULL;

   for( ptr = scgi_files; *ptr != NULL; ++ptr )
      if ( ptr == last )
         break;

   if ( ptr != last )
      return NULL;

   for( ptr += 4; *ptr != NULL; ptr += 5 )
   {
      ptr1 = *ptr;
      ptr2 = file;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( ! *ptr1 && ! *ptr2 )
         return ( ptr + 1 );
   }

   return NULL;
}

char **scgi_get_files()
{
   return scgi_files;
}

char *scgi_get_cookie( char *name )
{
   char **ptr;

   if (( ptr = scgi_lookup_list( scgi_cookies, name )) == NULL )
      return NULL;

   return *ptr;
}

char **scgi_get_cookies()
{
   return scgi_cookies;
}

void scgi_set_name( char *name )
{
   char *ptr1, *ptr2;
   int len;

   if ( name == NULL )
      return;

   scgi_app_name = name;

   for( len = 0, ptr1 = name; *ptr1; ++ptr1 )
      ++len;

   ptr2 = scgi_pidfile = scgi_memory( 14 + len );

   for ( ptr1 = "/var/run/"; *ptr1; )
      *ptr2++ = *ptr1++;

   for ( ptr1 = name; *ptr1; )
      *ptr2++ = *ptr1++;

   for ( ptr1 = ".pid"; *ptr1; )
      *ptr2++ = *ptr1++;

   *ptr2 = '\0';
}

void scgi_service_connections()
{
   for( ; ; )
   {
      scgi_accept_connection();

      if ( scgi_read_header() )
         continue;

      if ( scgi_read_body() )
         continue;

      scgi_request_handler( scgi_conn );
      scgi_free_data();
   }
}

void scgi_init_data()
{
   int i;

   if (( scgi_busymap = mmap( 0, scgi_max_workers, PROT_WRITE | PROT_READ,
                              MAP_ANON | MAP_SHARED, -1, 0 )) == MAP_FAILED )
   {
      syslog( LOG_ERR, "mmap(): %m" );
      exit( 1 );
   }

   scgi_workers = scgi_memory( sizeof( pid_t ) * scgi_max_workers );

   for( i = 0; i < scgi_max_workers; ++i )
   {
      scgi_busymap[ i ] = EMPTY;
      scgi_workers[ i ] = 0;
   }

   for( i = 0; i <= MAX_PARAM; ++i )
      scgi_params[ i ] = NULL;

   for( i = 0; i <= MAX_COOKIE; ++i )
      scgi_cookies[ i ] = NULL;

   for( i = 0; i <= MAX_FILE; ++i )
      scgi_files[ i ] = NULL;
}

void scgi_fork_worker( int i )
{
   pid_t pid;

   /*
    * Prevents a race condition.  Setting the worker not busy here will
    * prevent the loop that counts idle processes in scgi_manage_workers()
    * from failing to count the new worker process before it has changed its
    * its initial busymap entry from EMPTY to NOT_BUSY.  Otherwise, the
    * server will fork too many processes at startup if scgi_worker_init_func()
    * is slow to complete.
    */

   scgi_busymap[ i ] = NOT_BUSY;

   switch(( pid = fork()))
   {
      case -1:
         syslog( LOG_ERR, "fork(): %m" );
         scgi_busymap[ i ] = EMPTY;
         exit( 1 );

      case 0:
         free( scgi_workers );
         scgi_workers = NULL;

         signal( SIGSEGV, SIG_DFL );
         signal( SIGBUS, SIG_DFL );

         scgi_worker_no = i;
         scgi_worker_init_func();
         scgi_service_connections();
         exit( 0 );

      default:
         scgi_workers[ i ] = pid;
   }
}

void scgi_set_periodic( void ( *func )(), int interval )
{
   scgi_periodic_func = func;

   if ( interval < 0 )
      interval = 0;

   scgi_interval = interval * ( SECOND / NANOSECONDS );
   scgi_counter = 0;
}

void scgi_manage_workers()
{
   pid_t pid;
   int i, t, idle, last_idle, half, two_seconds;
   struct timespec tm;

   t = 0;
   if ( !( half = scgi_min_workers / 2 ))
      half = 1;

   tm.tv_sec = 0;
   tm.tv_nsec = NANOSECONDS;

   two_seconds = ( SECOND / NANOSECONDS ) * 2;

   for( ; ; )
   {
      nanosleep( &tm, NULL );

      if ( scgi_killed )
         scgi_kill_workers_and_exit();

      ++t;

      if ( scgi_interval && scgi_periodic_func != NULL && ++scgi_counter == scgi_interval )
      {
         scgi_periodic_func();
         scgi_counter = 0;
      }

      idle = 0;
      last_idle = -1;

      for( i = 0; i < scgi_max_workers; ++i )
         if ( scgi_busymap[ i ] == NOT_BUSY )
         {
            last_idle = i;
            ++idle;
         }

      if ( t == two_seconds )
      {
         t = 0;

         if ( idle > scgi_min_workers )
         {
            kill( scgi_workers[ last_idle ], SIGTERM );
            --idle;
         }
      }

      while(( pid = waitpid( -1, NULL, WNOHANG )) > 0 )
      {
         for( i = 0; i < scgi_max_workers; ++i )
            if ( pid == scgi_workers[ i ] )
            {
               scgi_workers[ i ] = 0;
               scgi_busymap[ i ] = EMPTY;
               break;
            }
      }

      if ( idle < half )
      {
         for( i = 0; i < scgi_max_workers; ++i )
            if ( ! scgi_workers[ i ] )
            {
               scgi_fork_worker( i );

               if ( ++idle == scgi_min_workers )
                  break;
            }
      }
   }
}

void scgi_set_sigterm_intr()
{
   struct sigaction sigact;

   sigact.sa_handler = scgi_signal_handler;
   sigemptyset( &sigact.sa_mask );
   sigact.sa_flags = 0;

   if ( sigaction( SIGTERM, &sigact, NULL ) < 0 )
      syslog( LOG_ERR, "dombey: sigaction: %s\n", strerror( errno ));
}

void scgi_set_signals()
{
   int *iptr;
   int sigs[] = {
      SIGPIPE, SIGHUP, SIGQUIT, SIGUSR1, SIGUSR2, SIGALRM, SIGINT, SIGTSTP, -1
   };

   /*
    * We ignore everything but SIGTERM, SIGBUS, and SIGSEGV when not testing.
    */

   signal( SIGTERM, scgi_signal_handler );

   for( iptr = sigs; *iptr > 0; ++iptr )
      signal( *iptr, ( scgi_testing ? scgi_signal_handler : SIG_IGN ));
}

void scgi_fork_workers()
{
   int i;

   for( i = 0; i < scgi_min_workers; ++i )
      scgi_fork_worker( i );
}

char *scgi_base64_decode( char *input )
{
   unsigned int i, pad;
   char buff[ 4 ], *ptr;
   static struct scgi_string *s = NULL;

   if ( s == NULL )
      s = scgi_make_string();
   else
   {
      s->free += s->used;
      s->used = 0;
      s->top = s->str;
   }

   for( ptr = input; *ptr; )
   {
      pad = 0;

      for( i = 0; i < 4; ++i, ++ptr )
      {
         if ( *ptr >= 'A' && *ptr <= 'Z' )
            buff[ i ] = *ptr - 65;
         else if ( *ptr >= 'a' && *ptr <= 'z' )
            buff[ i ] = *ptr - 97 + 26;
         else if ( *ptr >= '0' && *ptr <= '9' )
            buff[ i ] = *ptr - 48 + 52;
         else if ( *ptr == '+' )
            buff[ i ] = 62;
         else if ( *ptr == '/' )
            buff[ i ] = 63;
         else if ( *ptr == '=' )
         {
            buff[ i ] = 0;
            ++pad;
         }
         else
            return NULL;
      }

      buff[ 0 ] <<= 2;
      buff[ 0 ] += ( buff[ 1 ] & 0x30 ) >> 4;
      buff[ 1 ] <<= 4;
      buff[ 1 ] += ( buff[ 2 ] & 0x3c ) >> 2;
      buff[ 2 ] <<= 6;
      buff[ 2 ] += buff[ 3 ];

      pad = 3 - pad;

      for( i = 0; i < pad; ++i )
         STRING_APPEND( s, buff[ i ] )
   }

   scgi_inlen = s->used;
   return s->str;
}

void scgi_encrypt_files( int on )
{
   scgi_encrypting = on;
}

int main( int argc, char **argv )
{
   ERR_load_crypto_strings();

   scgi_set_options( argc, argv );
   scgi_set_signals();

   scgi_init_func();
   openlog( scgi_app_name, LOG_PID, LOG_DAEMON );
   scgi_logging = 1;

   if ( !scgi_testing )
      scgi_become_daemon();

   scgi_start_listening();
   scgi_change_identity();
   scgi_init_data();

   if ( scgi_testing )
   {
      scgi_worker_init_func();
      scgi_service_connections();
   }
   else
   {
      scgi_fork_workers();
      scgi_manage_workers();
   }

   return 0;
}
